/* -*- tab-width: 8; -*- */
/*
 * Virtual Machine version 3
 */

#ifdef VM3_PROFILE
#include <time.h>
#include <sys/times.h>
#endif 

/* slot offset for SCM_ENVFRAME */
#define ENV_NEXT		0
#define ENV_NSLOTS		1
#define ENV_SLOT		2
#define ENV_HEADER_SLOTS	2

#define CONTEXT		struct _CONTEXT
CONTEXT {
  CONTEXT	*next;
  SCM_ENVFRAME	*env;
  SOBJ		*ip;
};

/* slot offset for CONTEXT */
#define CONT_NEXT		0
#define CONT_ENV		1
#define CONT_IP			2
#define CONT_TOTAL_SLOTS	3

#define SCM_DICT_MAX_DEPTH	4

#define SCM_DICT	struct _SCM_DICT
#define SCM_DICT_ENTRY	struct _SCM_DICT_ENTRY

SCM_DICT_ENTRY {
  SCM_DICT_ENTRY *next;
  void *key;
  void *value;
};

SCM_DICT {
  int  	size;			/* number of entries in dict array */
  int	nkeys;			/* current number of keys stored */
  int 	maxkeys;		/* max depth */
  SCM_DICT_ENTRY *(*newkey)(); 	/* key,value */
  void	(*setkey)();		/* DICT_ENTRY *, key */
  Uint	(*hash)();		/* hash function */
  int	(*cmp)();		/* compare keys: result like strcmp */
  SCM_DICT_ENTRY **entry;	/* entries in dictionnary */
};


#ifdef VM3_PROFILE_HIRES
#define CLOCK_T		long long
#define CLOCK_FQ	(1 / profile_cycle_period)		
#else
#define CLOCK_T		clock_t
#define CLOCK_FQ	CLOCKS_PER_SEC
#endif

typedef struct {
  void *na,*start, *end;	/* start, end address of an opcode */
  char *name;			/* name of proc */
  char arity;			/* number of argument expected on stack */
  char follow;			/* number of words following the code */
  char term;			/* does this opcode terminate the engine */
  void  (*jcomp)();		/* jit compiler for this opcode */
#ifdef VM3_PROFILE
  long count;			/* profile counter*/
  CLOCK_T spent;		/* time spent */
  CLOCK_T start_clock;		/* start clock */
#endif /*VM3_PROFILE*/
} SCM_VM3_PRIM;

enum SCM_VM3_DOERS {
  SCM_VM3_EXECUTE = 0,
  SCM_VM3_MCODE_EXECUTE,
  SCM_VM3_INIT,
  SCM_VM3_ASSEMBLE,
  SCM_VM3_NASSEMBLE,
  SCM_VM3_DISASSEMBLE,
  SCM_VM3_MAX_DOERS
};

#define FAST_DISPATCH

#define SCM_CONT_NSLOTS	(sizeof(SCM_ContFrame)/sizeof(SOBJ))
#define SCM_ENV0_NSLOTS	2

enum SCM_VM3_OPCODES {
#include "vm3-ops.h"
};

#define ASMV		__asm__ __volatile__

/*** Register affectation */
#ifdef __i386__
#define TOSREG 	asm("%ecx")
#define IPREG	asm("%esi")
#define SPREG	asm("%edi")
#define EREG 	asm("%ebx")

/* JIT register maping */
#define VM_TMP	_EAX
#define VM_TOS	_ECX
#define VM_ENV	_EBX
#define VM_TMP2 _EDX
#define VM_SP	_EDI
#define VM_IP	_ESI

/* define this when env is is registers */
#define HAS_ENV_REG

/* Load VM_TMP (%eax) with VM REGISTERS */
#define LOAD_ENV()	ASMV("movl %0, %%eax" : : "g" (env))
#define LOAD_CONT()	ASMV("movl %0, %%eax" : : "g" (cont))

#define SAFE(x) { SOBJ save=TOS, x; TOS=save; }
#endif /* __i386__ */

/* default register affectation (none) for TOS, IP, SP, CONT and ENV */
#ifndef TOSREG
#define TOSREG 
#endif

#ifndef IPREG
#define IPREG 
#endif

#ifndef SPREG
#define SPREG 
#endif

#ifndef CREG
#define CREG
#endif

#ifndef EREG
#define EREG
#endif


/*** Stack definition ***/
/*
 * Stack grows up and TOS caches the top of the stack. SP points to
 * the current * top of stack
 *
 * NOTE: this macros are to be used strictly in the vm3 engine
 */
#define sclear()    { sp = vm->stack_base;  TOS=NULL; }
#define supdate()   { *sp = TOS;  }
#define spush(x)    { *sp = TOS; sp++; TOS = (x); }
#define spop(x)	    { (x) = TOS; sp--; TOS = *sp; }
#define sdrop()	    { sp--; TOS = *sp; }
#define sresync()   { vm->reg.sp = sp; }
#define sdepth()    (sp - vm->stack_base)

#ifdef FAST_DISPATCH

#define START_VMCODE	PROFILE_INIT();
#define END_VMCODE

#define NEXT 		goto *(*ip++)

#ifdef VM3_PROFILE
#define PROFILE_INIT()		profile_init(symbol)
#define PROFILE_DUMP()		profile_dump(symbol)
#define PROFILE_START(name)	profile_start(symbol + SCM_OP_##name##)
#define PROFILE_STOP(name)	profile_stop(symbol + SCM_OP_##name##)


#else  /*VM3_PROFILE*/
#define PROFILE_INIT()	
#define PROFILE_DUMP()	
#define PROFILE_START(name)
#define PROFILE_STOP(name)

#endif /*VM3_PROFILE*/

#define ALIGN()		ASMV(".align 8\n")

#ifndef VM3_DEBUG
#define INSTR(name)	a_##name##: ALIGN(); l_##name##: PROFILE_START(name);
#define END(name)	e_##name##: PROFILE_STOP(name); NEXT

#else	/* VM3_DEBUG */

#define DUMP(what,name)	{ \
  supdate(); fprintf(stderr, "VM3:%s:%s\nVM3:cont=%p env=%p ip=%p sp=%p\n",\
	                     what,#name,cont,env,ip,sp); \
  scm_vm3_dump_stack(sp); TOS=*sp; }

#define INSTR(name)	a_##name##: ALIGN(); l_##name##: DUMP("enter",name); PROFILE_START(name);
#define END(name)	e_##name##: PROFILE_STOP(name); NEXT

#endif	/* VM3_DEBUG */

#else
#define START_VMCODE	PROFILE_INIT(); switch(*ip++) {
#define END_VMCODE	}

#define NEXT		break
#define INSTR(name)	case SCM_OP_##name##:

#endif

/* declaration of a new opcode in the SCM_VM3_PRIM structure */
#define DECL_REF(n) &&a_##n,&&l_##n,&&e_##n, "%"#n

/* #define DECL(name,arity,follow,term) {DECL_REF(name),arity,follow,term} */

/* args should be: arity, follow, term, ... */
#define DECL(name, args...) {DECL_REF(name), ##args}

#define PRIM(lab,name,narg) 	INSTR(lab)
#define PRIMV(lab,name)		INSTR(lab)

#define END_PRIM(lab)		END(lab)
#define RETURN(x)		TOS = (x);

/* note: x can be a C function call. Don't change env context before calling */
#define VRETURN(x) { SOBJ res=(x); sp=(SOBJ*)env; env=env->next; TOS=res; }
#define NARGS	   ((sp - env->slot)+1)

#define REQUIRES(x,t,msg) \
   if(x == NULL||SCM_INUMP(x)||((SOBJ)x)->type!=t){SCM_ERR(msg,x);}

/*** Proto ***/
extern SCM_DICT		*scm_prim_name_dict;
extern SCM_DICT		*scm_prim_addr_dict;

/*** CFUNC PROTO ***/
Uint        scm_dict_string_hash(unsigned char *s);
SCM_DICT_ENTRY *scm_dict_string_newkey(char *key, void *value);
void        scm_dict_string_setkey(SCM_DICT_ENTRY *de, void *value);
void        scm_dict_set_string_type(SCM_DICT *dict);
Uint        scm_dict_addr_hash(void *p);
SCM_DICT_ENTRY *scm_dict_addr_newkey(void *addr, void *value);
void        scm_dict_addr_setkey(SCM_DICT_ENTRY *de, void *value);
int         scm_dict_addr_cmp(void *x, void *y);
void        scm_dict_set_addr_type(SCM_DICT *dict);
SCM_DICT    *scm_dict_new();
void        scm_dict_dump(SCM_DICT *dict);
SCM_DICT_ENTRY *scm_dict_search(SCM_DICT *dict, void *key);
SCM_DICT_ENTRY *scm_dict_set(SCM_DICT *dict, void *key, void *value);
void        scm_vm3_dump_env(SCM_ENVFRAME *e);
void        scm_vm3_dump_cont(CONTEXT *c);
void        scm_vm3_dump_stack(SOBJ *sp);
SCM_VM3_PRIM *scm_vm3_search_name(SCM_VM3_PRIM *t, char *name);
SCM_VM3_PRIM *scm_vm3_search_addr(SCM_VM3_PRIM *t, void *addr);
void        scm_vm3_engine_init(SCM_VMD *vm, SCM_VM3_PRIM *t);
SOBJ        scm_vm3_make_closure(SCM_VMD *vm, SCM_ENVFRAME *env, SOBJ *sp, SOBJ proc);
void        scm_vm3(SCM_VMD *vm);
SOBJ        scm_vm3_init();
SOBJ        scm_vm3_asm(SOBJ code);
SOBJ        scm_vm3_nasm(SOBJ code);
SOBJ        scm_vm3_dis(SOBJ code);
SOBJ        scm_vm3_exec(SOBJ code);
SOBJ        scm_make_proc(SOBJ v, SOBJ env, SOBJ nargs, SOBJ optarg);
SOBJ        scm_mcode_dis(SOBJ start_addr, SOBJ length);
void        scm_init_vm3();
